import os, sys
import xam
from ivmdump import insnlist, insntable, stackpushes, chainable
from struct import unpack


IGNORE_INSNS = ('assertdepth', 'dynamicfreq')

def dump(data):
    l = len(data)
    p = 0
    results = []
    result = []
    freq = 1
    while p < l:
        mode = insnlist[ord(data[p])]
        if mode.opcode not in insntable:
            break
        p += mode.unpacksize + 1
        if p > l:
            break
        args = mode.getargs(data, p-mode.unpacksize)
        for insn in mode.insns:
            if insn[0] in IGNORE_INSNS:
                if insn[0] == 'dynamicfreq':
                    freq = args[0]
                continue
            a = len(insn)-1
            if a:
                txt = '%s(%s)' % (insn[0], ','.join(map(str,args[:a])))
                del args[:a]
            else:
                txt = insn[0]
            result.append(txt)
        if mode.opcode not in chainable:
            results.append((freq, result))
            result = []
    results.append((freq, result))
    return results


def main(DIRECTORY):
    filename = os.path.join(DIRECTORY, 'psyco.dump')
    if not os.path.isfile(filename) and os.path.isfile(DIRECTORY):
        filename = DIRECTORY
        DIRECTORY = os.path.dirname(DIRECTORY)
    outfilename = filename + '.ivm'
    if os.path.isfile(filename):
        codebufs = xam.readdump(filename)
        f = open(outfilename, 'w')
        for codebuf in codebufs:
            if codebuf.data:
                data, addr, next, key = codebuf.splitheader()
                for freq, lst in dump(data):
                    if len(lst) > 1:
                        print >> f, 'psycodump(%d, [%s]).' % (freq,
                                                              ', '.join(lst))
        f.close()
    elif not os.path.isfile(outfilename):
        print >> sys.stderr, filename, "not found."
        sys.exit(1)
    else:
        print >> sys.stderr, "reusing text dump from", outfilename
    return outfilename


if __name__ == '__main__':
    if len(sys.argv) <= 1:
        print "Usage: python ivmextract.py <directory>"
        print "  psyco.dump is loaded from the <directory>."
        sys.exit(2)
    for dir in sys.argv[1:]:
        print "'%s'." % main(dir)



###########################################################################
# 
#  Psyco class support module.
#   Copyright (C) 2001-2002  Armin Rigo et.al.

"""Psyco class support module.

'psyco.classes.psyobj' is an alternate Psyco-optimized root for classes.
Any class inheriting from it or using the metaclass '__metaclass__' might
get optimized specifically for Psyco. It is equivalent to call
psyco.bind() on the class object after its creation.

Importing everything from psyco.classes in a module will import the
'__metaclass__' name, so all classes defined after a

       from psyco.classes import *

will automatically use the Psyco-optimized metaclass.
"""
###########################################################################

__all__ = ['psyobj', 'psymetaclass', '__metaclass__']


from _psyco import compacttype
import core
from types import FunctionType

class psymetaclass(compacttype):
    "Psyco-optimized meta-class. Turns all methods into Psyco proxies."

    def __new__(cls, name, bases, dict):
        bindlist = dict.get('__psyco__bind__')
        if bindlist is None:
            bindlist = [key for key, value in dict.items()
                        if isinstance(value, FunctionType)]
        for attr in bindlist:
            dict[attr] = core.proxy(dict[attr])
        return super(psymetaclass, cls).__new__(cls, name, bases, dict)

psyobj = psymetaclass("psyobj", (), {})
__metaclass__ = psymetaclass



###########################################################################
# 
#  Psyco main functions.
#   Copyright (C) 2001-2002  Armin Rigo et.al.

"""Psyco main functions.

Here are the routines that you can use from your applications.
These are mostly interfaces to the C core, but they depend on
the Python version.

You can use these functions from the 'psyco' module instead of
'psyco.core', e.g.

    import psyco
    psyco.log('/tmp/psyco.log')
    psyco.profile()
"""
###########################################################################

import _psyco
import types, new
from support import *


# Default charge profiler values
default_watermark     = 0.09     # between 0.0 (0%) and 1.0 (100%)
default_halflife      = 0.5      # seconds
default_pollfreq_profile    = 20       # Hz
default_pollfreq_background = 100      # Hz -- a maximum for sleep's resolution
default_parentframe   = 0.25     # should not be more than 0.5 (50%)


def full(memory=None, time=None, memorymax=None, timemax=None):
    """Compile as much as possible.

Typical use is for small scripts performing intensive computations
or string handling."""
    import profiler
    p = profiler.FullCompiler()
    p.run(memory, time, memorymax, timemax)


def profile(watermark   = default_watermark,
            halflife    = default_halflife,
            pollfreq    = default_pollfreq_profile,
            parentframe = default_parentframe,
            memory=None, time=None, memorymax=None, timemax=None):
    """Turn on profiling.

The 'watermark' parameter controls how easily running functions will
be compiled. The smaller the value, the more functions are compiled."""
    import profiler
    p = profiler.ActivePassiveProfiler(watermark, halflife,
                                       pollfreq, parentframe)
    p.run(memory, time, memorymax, timemax)


def background(watermark   = default_watermark,
               halflife    = default_halflife,
               pollfreq    = default_pollfreq_background,
               parentframe = default_parentframe,
               memory=None, time=None, memorymax=None, timemax=None):
    """Turn on passive profiling.

This is a very lightweight mode in which only intensively computing
functions can be detected. The smaller the 'watermark', the more functions
are compiled."""
    import profiler
    p = profiler.PassiveProfiler(watermark, halflife, pollfreq, parentframe)
    p.run(memory, time, memorymax, timemax)


def runonly(memory=None, time=None, memorymax=None, timemax=None):
    """Nonprofiler.

XXX check if this is useful and document."""
    import profiler
    p = profiler.RunOnly()
    p.run(memory, time, memorymax, timemax)


def stop():
    """Turn off all automatic compilation.  bind() calls remain in effect."""
    import profiler
    profiler.go([])


def log(logfile='', mode='w', top=10):
    """Enable logging to the given file.

If the file name is unspecified, a default name is built by appending
a 'log-psyco' extension to the main script name.

Mode is 'a' to append to a possibly existing file or 'w' to overwrite
an existing file. Note that the log file may grow quickly in 'a' mode."""
    import profiler, logger
    if not logfile:
        import os
        logfile, dummy = os.path.splitext(sys.argv[0])
        if os.path.basename(logfile):
            logfile += '.'
        logfile += 'log-psyco'
    if hasattr(_psyco, 'VERBOSE_LEVEL'):
        print >> sys.stderr, 'psyco: logging to', logfile
    # logger.current should be a real file object; subtle problems
    # will show up if its write() and flush() methods are written
    # in Python, as Psyco will invoke them while compiling.
    logger.current = open(logfile, mode)
    logger.print_charges = top
    profiler.logger = logger
    logger.writedate('Logging started')
    cannotcompile(logger.psycowrite)
    _psyco.statwrite(logger=logger.psycowrite)


def bind(x, rec=None):
    """Enable compilation of the given function, method, or class object.

If C is a class (or anything with a '__dict__' attribute), bind(C) will
rebind all functions and methods found in C.__dict__ (which means, for
classes, all methods defined in the class but not in its parents).

The optional second argument specifies the number of recursive
compilation levels: all functions called by func are compiled
up to the given depth of indirection."""
    if isinstance(x, types.MethodType):
        x = x.im_func
    if isinstance(x, types.FunctionType):
        if rec is None:
            x.func_code = _psyco.proxycode(x)
        else:
            x.func_code = _psyco.proxycode(x, rec)
        return
    if hasattr(x, '__dict__'):
        funcs = [o for o in x.__dict__.values()
                 if isinstance(o, types.MethodType)
                 or isinstance(o, types.FunctionType)]
        if not funcs:
            raise error, ("nothing bindable found in %s object" %
                          type(x).__name__)
        for o in funcs:
            bind(o, rec)
        return
    raise TypeError, "cannot bind %s objects" % type(x).__name__


def unbind(x):
    """Reverse of bind()."""
    if isinstance(x, types.MethodType):
        x = x.im_func
    if isinstance(x, types.FunctionType):
        try:
            f = _psyco.unproxycode(x.func_code)
        except error:
            pass
        else:
            x.func_code = f.func_code
        return
    if hasattr(x, '__dict__'):
        for o in x.__dict__.values():
            if (isinstance(o, types.MethodType)
             or isinstance(o, types.FunctionType)):
                unbind(o)
        return
    raise TypeError, "cannot unbind %s objects" % type(x).__name__


def proxy(x, rec=None):
    """Return a Psyco-enabled copy of the function.

The original function is still available for non-compiled calls.
The optional second argument specifies the number of recursive
compilation levels: all functions called by func are compiled
up to the given depth of indirection."""
    if isinstance(x, types.FunctionType):
        if rec is None:
            code = _psyco.proxycode(x)
        else:
            code = _psyco.proxycode(x, rec)
        return new.function(code, x.func_globals, x.func_name)
    if isinstance(x, types.MethodType):
        p = proxy(x.im_func, rec)
        return new.instancemethod(p, x.im_self, x.im_class)
    raise TypeError, "cannot proxy %s objects" % type(x).__name__


def unproxy(proxy):
    """Return a new copy of the original function of method behind a proxy.
The result behaves like the original function in that calling it
does not trigger compilation nor execution of any compiled code."""
    if isinstance(proxy, types.FunctionType):
        return _psyco.unproxycode(proxy.func_code)
    if isinstance(proxy, types.MethodType):
        f = unproxy(proxy.im_func)
        return new.instancemethod(f, proxy.im_self, proxy.im_class)
    raise TypeError, "%s objects cannot be proxies" % type(proxy).__name__


def cannotcompile(x):
    """Instruct Psyco never to compile the given function, method
or code object."""
    if isinstance(x, types.MethodType):
        x = x.im_func
    if isinstance(x, types.FunctionType):
        x = x.func_code
    if isinstance(x, types.CodeType):
        _psyco.cannotcompile(x)
    else:
        raise TypeError, "unexpected %s object" % type(x).__name__


def dumpcodebuf():
    """Write in file psyco.dump a copy of the emitted machine code,
provided Psyco was compiled with a non-zero CODE_DUMP.
See py-utils/httpxam.py to examine psyco.dump."""
    if hasattr(_psyco, 'dumpcodebuf'):
        _psyco.dumpcodebuf()


###########################################################################
# Psyco variables
#   error         * the error raised by Psyco
#   warning       * the warning raised by Psyco
#   __in_psyco__  * a new built-in variable which is always zero, but which
#                     Psyco special-cases by returning 1 instead. So
#                     __in_psyco__ can be used in a function to know if
#                     that function is being executed by Psyco or not.



from __future__ import generators
import os, sys, random
import psyco


def do_test_1(objects):
    d = [{} for i in range(20)]
    s = [psyco.compact() for i in range(20)]
    attrnames = list('abcdefghijklmnopqrstuvwxyz')
    for j in range(5000):
        i = random.randrange(0, 20)
        attr = random.choice(attrnames)
        if random.randrange(0, 2):
            if attr in d[i]:
                if random.randrange(0,5) == 3:
                    delattr(s[i], attr)
                    del d[i][attr]
                else:
                    assert d[i][attr] == getattr(s[i], attr)
            else:
                try:
                    getattr(s[i], attr)
                except AttributeError:
                    pass
                else:
                    raise AssertionError, attr
        else:
            obj = random.choice(objects)
            setattr(s[i], attr, obj)
            d[i][attr] = obj
    for i in range(20):
        d1 = {}
        for attr in attrnames:
            try:
                d1[attr] = getattr(s[i], attr)
            except AttributeError:
                pass
        assert d[i] == d1

def do_test(n, do_test_1=do_test_1):
    random.jumpahead(n*111222333444555666777L)
    N = 1
    TAIL = 'lo'
    objects = [None, -1, 0, 1, 123455+N, -99-N,
               'hel'+TAIL, [1,2], {(5,): do_test}, 5.43+0.01*N, xrange(5)]
    do_test_1(objects)
    for o in objects[4:]:
        #print '%5d  -> %r' % (sys.getrefcount(o), o)
        assert sys.getrefcount(o) == 4

psyco.cannotcompile(do_test)


def subprocess_test(n):
    sys.stdout.flush()
    childpid = os.fork()
    if not childpid:
        do_test(n)
        sys.exit(0)
    childpid, status = os.wait()
    return os.WIFEXITED(status) and os.WEXITSTATUS(status) == 0

##def test_compact_stress(repeat=20):
##    for i in range(repeat):
##        yield subprocess_test, i

# ____________________________________________________________

def read_x(k):
    return k.x + 1

def read_y(k):
    return k.y

def read_z(k):
    return k.z

psyco.bind(read_x)
psyco.bind(read_y)

def pcompact_test():
    k = psyco.compact()
    k.x = 12
    k.z = None
    k.y = 'hello'
    print read_x(k)
    print read_y(k)
    print read_z(k)
    #psyco.dumpcodebuf()

def pcompact_creat(obj):
    base = sys.getrefcount(obj)
    items = []
    for i in range(11):
        k = psyco.compact()
        k.x = (0,i,i*2)
        k.y = i+1
        k.z = None
        k.t = obj
        k.y = i+2
        k.y = i+3
        print k.x, k.y, k.z, k.t
        items.append(k)
        del k
    print sys.getrefcount(obj) - base
    del items[:]
    print sys.getrefcount(obj) - base

psyco.bind(pcompact_creat)

def pcompact_modif(obj):
    base = sys.getrefcount(obj)
    for i in range(21):
        k = psyco.compact()
        #k.x = i+1
        #k.y = (i*2,i*3,i*4,i*5,i*6,i*7)
        k.x = obj
        print k.x,
        print sys.getrefcount(obj) - base,
        k.x = i+1
        print sys.getrefcount(obj) - base,
        print k.x
        #k.x = len
        #k.x = i+2
        #print k.x, k.y
    print sys.getrefcount(obj) - base

psyco.bind(pcompact_modif)

# ____________________________________________________________

class Rect(psyco.compact):
    def __init__(self, w, h):
        self.w = w
        self.h = h
    def getarea(self):
        return self.w * self.h

def test_rect():
    assert Rect(10, 12).getarea() == 120
    assert Rect(0.5, 2.5).getarea() == 1.25
    assert Rect([1,2,3], 2).getarea() == [1,2,3,1,2,3]

class ClassWithNoInit(psyco.compact):
    pass

def test_init_arguments():
    import py
    def f1(): Rect()
    def f2(): psyco.compact(12)
    def f3(): ClassWithNoInit(21)
    py.test.raises(TypeError, f1)
    if sys.version >= (2, 3):
        py.test.raises(TypeError, f2)
        py.test.raises(TypeError, f3)
    py.test.raises(TypeError, psyco.proxy(f1))
    if sys.version >= (2, 3):
        py.test.raises(TypeError, psyco.proxy(f2))
        py.test.raises(TypeError, psyco.proxy(f3))

def test_special_attributes():
    missing = object()
    r = Rect(6, 7)
    assert r.__members__ == ['w', 'h']
    assert r.__dict__.items() == [('w', 6), ('h', 7)]
    assert r.__dict__ == {'w': 6, 'h': 7}
    assert {'w': 6, 'h': 7} == r.__dict__
    assert list(r.__dict__) == ['w', 'h']
    del r.__dict__['w']
    assert getattr(r, 'w', missing) is missing
    assert r.h == 7
    assert r.__members__ == ['h']
    assert r.__dict__.items() == [('h', 7)]
    assert r.__dict__ == {'h': 7}
    assert {'h': 7} == r.__dict__
    assert list(r.__dict__) == ['h']
    del r.h
    assert r.__members__ == []
    assert getattr(r, 'w', missing) is missing
    assert getattr(r, 'h', missing) is missing

def test_inheritance():
    class X(psyco.compact):
        pass
    class Y(psyco.compact):
        pass
    class Z(X):
        pass
    x = X()
    x.a = 5
    assert [s for s in dir(x) if not s.startswith('__')] == ['a']
    x.__class__ = Y
    assert type(x) is Y
    assert x.__class__ is Y
    assert Y.__bases__ == (psyco.compact,)
    assert Z.__bases__ == (X,)
    if sys.version >= (2,3):   # can't assign to __bases__ in Python 2.2
        Z.__bases__ = (Y,)
        assert Z.__bases__ == (Y,)

psyco.cannotcompile(test_inheritance)  # because of type mutation

def test_data_descr():
    global done
    done = []
    class X(psyco.compact):
        def g(self): done.append('g')
        def s(self, value): done.append(value)
        def d(self): done.append('d')
        a = property(g,s,d)
    x = X()
    x.__dict__['a'] = 'this is hidden'
    r = x.a
    assert r is None
    x.a = 123
    del x.a
    del x.a
    assert done == ['g', 123, 'd', 'd']
    assert x.__dict__ == {'a': 'this is hidden'}
    x.__dict__ = {'a': 'this too'}
    assert x.__dict__ == {'a': 'this too'}
    assert done == ['g', 123, 'd', 'd']

def test_ass_dict():
    missing = object()
    x = psyco.compact()
    x.a = 5
    assert x.__dict__ == {'a': 5}
    x.__dict__ = {'b': 6}
    assert x.b == 6
    assert getattr(x, 'a', missing) is missing
    assert x.__dict__ == {'b': 6}
    y = psyco.compact()
    y.__dict__ = x.__dict__
    assert y.__dict__ == {'b': 6}
    y.__dict__ = y.__dict__
    assert y.__dict__ == {'b': 6}

def test_with_psyco():
    yield psyco.proxy(test_rect)
    yield psyco.proxy(test_special_attributes)
    yield psyco.proxy(test_data_descr)

def test_compact_stress(repeat=20):
    for i in range(repeat):
        yield do_test, i

rect1 = Rect(0, 0)

def test_constant_obj():
    def f1():
        return rect1.w * rect1.h
    def f2(a):
        rect1.w = a
    psyco.bind(f1)
    psyco.bind(f2)
    rect1.w = 6
    rect1.h = 7
    res = f1()
    assert res == 42
    f2(4)
    res = f1()
    assert res == 28
    f2(0.5)
    res = f1()
    assert res == 3.5

# ____________________________________________________________

if __name__ == '__main__':
    import time; print "break!"; time.sleep(1)
    #subprocess_test(10)
    #pcompact_test()
    #pcompact_creat('hel' + 'lo')
    #pcompact_modif('hel' + 'lo')
    #test_constant_obj()
    psyco.proxy(test_rect)()
    #psyco.proxy(test_special_attributes)()
    #psyco.proxy(test_data_descr)()
    psyco.dumpcodebuf()



###########################################################################
# 
#  Psyco top-level file of the Psyco package.
#   Copyright (C) 2001-2002  Armin Rigo et.al.

"""Psyco -- the Python Specializing Compiler.

Typical usage: add the following lines to your application's main module,
preferably after the other imports:

try:
    import psyco
    psyco.full()
except ImportError:
    print 'Psyco not installed, the program will just run slower'
"""
###########################################################################


#
# This module is present to make 'psyco' a package and to
# publish the main functions and variables.
#
# More documentation can be found in core.py.
#


# Try to import the dynamic-loading _psyco and report errors
try:
    import _psyco
except ImportError, e:
    extramsg = ''
    import sys, imp
    try:
        file, filename, (suffix, mode, type) = imp.find_module('_psyco', __path__)
    except ImportError:
        ext = [suffix for suffix, mode, type in imp.get_suffixes()
               if type == imp.C_EXTENSION]
        if ext:
            extramsg = (" (cannot locate the compiled extension '_psyco%s' "
                        "in the package path '%s')" % (ext[0], '; '.join(__path__)))
    else:
        extramsg = (" (check that the compiled extension '%s' is for "
                    "the correct Python version; this is Python %s)" %
                    (filename, sys.version.split()[0]))
    raise ImportError, str(e) + extramsg

# Publish important data by importing them in the package
from support import __version__, error, warning, _getrealframe, _getemulframe
from support import version_info, __version__ as hexversion
from core import full, profile, background, runonly, stop, cannotcompile
from core import log, bind, unbind, proxy, unproxy, dumpcodebuf
from _psyco import setfilter
from _psyco import compact, compacttype



import sys, os
from struct import unpack, calcsize, error


class argtypes:
    char = "b", "0x%x"
    byte = code_t = "B", "0x%x"
    int  = "i", "0x%x"
    long = word_t = "l", "0x%x"
    def stack(t):
        if isinstance(t, tuple):
            return t[0], "[%d]"
        else:
            return "", "[%s]" % t, t
    def indirect(t):
        return t

fn = os.path.join(os.path.dirname(__file__), '../c/ivm/prolog/insns-table.py')
execfile(fn, argtypes.__dict__, globals())


class Mode:
    
    def __init__(self, opcode):
        self.opcode = opcode
        if opcode in insntable:
            self.insns = insntable[opcode]
        else:
            self.insns = [("<%d>" % opcode,)]
        if len(self.insns) == 1:
            self.singleinsn = self.insns[0][0]
        else:
            self.singleinsn = None
        self.stackpushes = stackpushes.get(opcode)
        self.unpackfmt = "="
        self.template = ""
        self.constantargs = []
        i = 0
        line = '%10x\t'
        for insn in self.insns:
            args = []
            for arg in insn[1:]:
                if isinstance(arg, tuple):
                    self.unpackfmt += arg[0]
                    args.append(arg[1])
                    if len(arg)>2:
                        self.constantargs.append((i, arg[2]))
                else:
                    args.append(str(arg))
                    self.constantargs.append((i, arg))
                i += 1
            line += '%-11s %s' % (insn[0], ', '.join(args))
            self.template += line
            line = '\n          \t'
        self.unpacksize = calcsize(self.unpackfmt)
        
    def dump(self, data, address, position):
        data = data[position:position+self.unpacksize]
        args = unpack(self.unpackfmt, data)
        result = self.template % ((address,) + args)
        return position+self.unpacksize, result.split('\n'), args

    def getargs(self, data, position):
        data = data[position:position+self.unpacksize]
        args = list(unpack(self.unpackfmt, data))
        for i, value in self.constantargs:
            args.insert(i, value)
        return args

insnlist = [Mode(opcode) for opcode in range(256)]


def dump(data, originaddr):
    l = len(data)
    if l>8 and data[-4:] == '\x00\x00\x00\x00':
        p, = unpack("l", data[-8:-4])
        queue = ["", "          (promotion chained list: 0x%x)" % p]
        l -= 4
    else:
        queue = []
    depth = None
    result = []
    p = 0
    try:
        while p < l:
            mode = insnlist[ord(data[p])]
            p, lines, args = mode.dump(data, originaddr+p, p+1)
            if depth is not None:
                if mode.stackpushes is None:
                    depth = None
                else:
                    depth += mode.stackpushes
                    lines[-1] = '%-40s [%d]' % (lines[-1], depth)
            if mode.singleinsn == 'assertdepth':
                asserteddepth = args[0]/4
                if depth is not None and asserteddepth != depth:
                    err = '************* assertion error **************'
                    lines.append(err)
                    print >> sys.stderr, err
                    print >> sys.stderr, originaddr
                else:
                    lines = [(s+' ')[:s.find('assertdepth')] for s in lines]
                depth = asserteddepth
            result += lines
    except error:
        while p < l:
            result.append("  %10x\t<%d>" % (originaddr+p, ord(data[p])))
            p += 1
    result += queue
    return result
